#!/usr/bin/env python

'''Read/write data in the C64 G64 format and translate it to a text-based
TRK format for easy editing. The TRK file (.trk) can be edited by hand or
processed in-place by specifying various arguments to this program.
Afterward, it can be converted back to a G64 file.

Examples:

Create a working file called "doc.trk", rotating tracks 2 and 3 right
by 207 bits before writing the output text file. This can be done in two
separate steps (creating and then processing).

  $ nibedit -t 2-3 -r 207 doc.g64 doc.trk

Rotate track 34.5 right by 2 bits and write the result back to the same
TRK file in place.

  $ nibedit -t 34.5 -r 2 --overwrite doc.trk

Convert a TRK file back into a G64 for testing. This will warn if the
"doc2.g64" file exists and would be overwritten.

  $ nibedit doc.trk doc2.g64

While converting an image, these options can be applied:

* Rotate entire track right any number of bits. Can be used to create
  track-to-track alignment, as well as bytewise alignment for emulators.

Future options not supported yet:

* Align start to first sync in image
* Align start to longest sync
* Align start to arbitrary byte pattern
* Align start to arbitrary sector number. Must have valid Commodore DOS
  headers.

Images are processed in a text file format (.trk), which can be modified
in arbitrary ways, including:

* Insert, change, delete GCR bytes
* Change SYNC length, delete SYNCs, add them, etc. Syncs are decoded and
  list their length in number of bits. Example: SYNC:40 is a normal
  5-byte sync.
* Data is dumped as raw GCR, two 5-byte chunks per line (10 bytes GCR).
  Decoded data is displayed at the end of the line, after a # mark.
  All data after the # mark is ignored when reading a TRK file.
  GCR symbols that are not 8-bits long are described in binary with
  a trailing '.'. However, these non-hex values can only appear at the
  end of each track, not in the middle. Example: "1001." is "9" in 4 GCR bits.
* Specify arbitrary density for any byte ranges. Density is either a single
  int value (whole track) or a series of (start, end, density) tuples.
  (future, not supported yet)
'''

import argparse, array, logging, os, struct, tempfile, time
from binascii import hexlify
from bitstring import BitStream

# Version for TRK file format, bump when it changes
TRKFILE_VERSION = 1
TRKFILE_SIG = 'GCR-TRACK-DATA'
TRK_HEADER = '%s: %s' % (TRKFILE_SIG, TRKFILE_VERSION) + '''
date: %(DATE)s
orig_filename: %(ORIG_FILENAME)s
track_count: %(TRACK_COUNT)s
'''

MAX_TRACKS = 42
MAX_HALFTRACKS = MAX_TRACKS * 2
SYNC_SYMBOL = '0b1111111111' # 10 or more set bits

# .g64 file constants
G64_SIG = 'GCR-1541'
G64_HEADER_SIZE = 12

# GCR conversion table
GCR_CHUNK_SIZE = 5 # bytes
GCR_TABLE_HIGH = (
    -1,   -1,   -1,   -1, -1,   -1,   -1,   -1,
    -1, 0x80, 0x00, 0x10, -1, 0xc0, 0x40, 0x50,
    -1,   -1, 0x20, 0x30, -1, 0xf0, 0x60, 0x70,
    -1, 0x90, 0xa0, 0xb0, -1, 0xd0, 0xe0,   -1,
)
GCR_TABLE_LOW = (
    -1,   -1,   -1,   -1, -1,   -1,   -1,   -1,
    -1, 0x08, 0x00, 0x01, -1, 0x0c, 0x04, 0x05,
    -1,   -1, 0x02, 0x03, -1, 0x0f, 0x06, 0x07,
    -1, 0x09, 0x0a, 0x0b, -1, 0x0d, 0x0e,   -1,
)

PROFILING = False


class GCRFile(list):
    '''Read/write C64 GCR files in .g64 format.
    For ease of editing, the intermediate format (.trk) is a text file
    that can then be converted to .g64 again.

    Internally, tracks are stored as a zero-indexed list (t1 is offset 0).
    However, they are referenced externally by track number (e.g.,
    GCRFile[2.5] is track 2.5, stored at offset 3 internally).

    Each track is composed of a tuple: a binary string and density info.
    If the density is an integer, it applies to the whole track.
    If it's a sequence, it's a set of (start_bit, end_bit, density_value).
    Bits indices start from 0 and refer to the bitstring index.
    The start/end values conform to Python slice behavior.

    Empty tracks have a density of -1 and a 0-length string.
    '''
    def __init__(self, filename=None):
        list.__init__(self)
        self.orig_filename = None
        self.filetype = None
        self.halftracks = False
        self.track_limit = 7928
        if filename:
            self.read(filename)

    def __getitem__(self, i):
        '''Use a 1-based index (float) since halftracks are numbered 1 - 42,
        in increments of 0.5'''
        if i == 0:
            raise IndexError('tracks are a 1-based index')
        return list.__getitem__(self, int(i * 2) - 2)

    def read(self, filename):
        'Read in a g64 or trk file, automatically probing type'
        fp = open(filename, 'rb')
        data = fp.read(1024)
        fp.seek(0)
        if data.find(TRKFILE_SIG) != -1:
            self.read_trkfile(fp)
        elif data.startswith(G64_SIG):
            self.read_g64(fp)
        else:
            raise ValueError('%s not a valid G64/TRK image' % filename)

    def write(self, name=None, filetype=None):
        '''Write out a g64 or trk file, choosing type from name if not
        specified.'''
        # Try various methods to choose the filetype.
        if name:
            ext = os.path.splitext(name)[1]
            if ext:
                ext_filetype = ext[1:].upper()
                if not filetype:
                    filetype = ext_filetype
                else:
                    if ext_filetype != filetype:
                        logging.warn(
                            'writing type %s even though file extension is %s',
                        filetype, ext)
        else:
            # Use original file's type if none other is specified
            if not filetype:
                filetype = self.filetype
        # If no name is specified or overwriting original, make a tempfile
        # for new file and write to it instead.
        try:
            if not name or name == self.orig_filename:
                final_name = self.orig_filename
                fd, name = tempfile.mkstemp(prefix='nibedit-tmp')
                os.close(fd)
        except NameError:
            raise ValueError('need a filename to write')
        # Write the proper type of file, based on the chosen/original format.
        if filetype == 'G64':
            self.write_g64(name)
        elif filetype == 'TRK':
            self.write_trkfile(name)
        else:
            raise ValueError('need a valid format: G64/TRK for %s' % name)
        # If we are overwriting the original, wait until we've successfully
        # written the new file and atomically rename it into place.
        try:
            if name != final_name:
                os.rename(name, final_name)
        except NameError: pass

    def read_g64(self, fp):
        'Read and parse a .g64 file'
        # Read initial header, number of tracks, and track size
        data = fp.read()
        if not data.startswith(G64_SIG):
            raise ValueError('%s not a valid G64 image' % fp.name)
        version, num_tracks, self.track_limit = struct.unpack('<BBH',
            data[8:12])
        if version != 0:
            logging.warn('version %d may not be valid', version)
        if num_tracks == 0 or num_tracks > MAX_HALFTRACKS:
            logging.warn('track count %d may not be valid', num_tracks)
        self.filetype = 'G64'
        self.halftracks = True
        self.orig_filename = fp.name
        offset = G64_HEADER_SIZE
        # Chop out the known sections: track offsets and density header data
        header_len = num_tracks * 4
        offset_data = data[offset:offset + header_len]
        offset += header_len
        density_data = data[offset:offset + header_len]
        # Walk through each track, us offset and get its track data
        for t in xrange(num_tracks):
            # Parse the track offset.
            # If 0, there is no track here (e.g., skipped halftracks)
            track_offset = struct.unpack('<I', offset_data[t*4:(t+1)*4])[0]
            if track_offset == 0:
                self += [(BitStream(), -1)]
                self.halftracks = False
                continue
            # Parse the initial density and track length values
            density_info = struct.unpack('<I', density_data[t*4:(t+1)*4])[0]
            track_len = struct.unpack('<H', data[track_offset:
                track_offset+2])[0]
            assert track_len <= self.track_limit, 'track size incorrect'
            track_offset += 2
            # Get the actual track data itself as a BitStream from bytes
            track_data = BitStream(
                bytes=data[track_offset:track_offset + track_len])
            # If not density itself, it's an offset to a per-byte density map
            if density_info >= 4:
                density = data[density_info:density_info + self.track_limit / 4]
                density_info = parse_density_map(density)
            self += [(track_data, density_info)]

    def write_g64(self, filename, vice_compat=True):
        '''Write out a .g64 file. The optional vice_compat flag specifies
        whether we should set track size to 7928 bytes and output exactly
        84 tracks (default enabled).
        '''
        fp = open(filename, 'wb')
        # VICE requires this, plus it never hurts to write all tracks
        if len(self) != MAX_HALFTRACKS:
            logging.warn('writing g64 with %d tracks, may not be compatible',
                len(self))
        if vice_compat:
            track_limit = 7928
        else:
            track_limit = max((x[0].len + 7) / 8 for x in self)
            track_limit = min(track_limit, 8192)
        # Header + density/track offset groups (4 bytes per track)
        track_offset = G64_HEADER_SIZE + (2 * MAX_HALFTRACKS * 4)
        tracks, densities, track_offsets = [], [], []
        for data, density in self:
            if vice_compat:
                if not isinstance(density, int):
                    raise ValueError('extended density not allowed for VICE')
            track_len = (data.len + 7) / 8
            if track_len > 0:
                # Each track offset is the full size + the 2-byte track length
                track_offsets.append(track_offset)
                track_offset += track_limit + 2
                # Pad out to track_limit and a byte boundary with 1's
                padding_len = (track_limit * 8) - data.len 
                padded_track = data + BitStream('0b' + padding_len * '1')
                tracks.append(struct.pack('<H', track_len) +
                    padded_track.tobytes())
            else:
                padded_track = data
                track_offsets.append(0)
                density = 0
            densities.append(struct.pack('<I', density))
        fp.write(G64_SIG + struct.pack('<BBH', 0, len(self), track_limit))
        fp.write(''.join(struct.pack('<I', off) for off in track_offsets))
        fp.write(''.join(densities))
        fp.write(''.join(tracks))
        fp.close()

    def read_trkfile(self, fp):
        lines = trk_data_lines(fp)
        # Read header
        sig_version = int(get_kv(TRKFILE_SIG, lines.next()))
        if sig_version != TRKFILE_VERSION:
            raise ValueError('unsupported trk file version: %s', sig_version)
        date = get_kv('date', lines.next())
        self.orig_filename = get_kv('orig_filename', lines.next())
        track_count = int(get_kv('track_count', lines.next()))
        # Clear any previous track data
        del self[:]
        track_data = []
        track_num = None
        self.halftracks = False
        density = -1
        for line in lines:
            # Parse track number and prepare to get track data
            if line.startswith('t'):
                track_bin = BitStream().join(track_data)
                if track_num:
                    self += [(track_bin, density)]
                track_num_new = float(get_kv('track', line))
                if track_num is not None:
                    delta = track_num_new - track_num
                    if delta == 0.5:
                        self.halftracks = True
                    else:
                        assert delta == 1.0
                        self += [(BitStream(), -1)]
                track_data = []
                density = -1
                track_num = track_num_new
            elif line.startswith('den'):
                density = int(get_kv('density', line))
                if track_num is None:
                    raise ValueError('density before track number line')
            else:
                # Handle sync bits
                if line.startswith('S'):
                    bits = int(get_kv('SYNC', line))
                    track_data.append(BitStream('0b' + bits * '1'))
                else:
                    if line.endswith('.'):
                        try:
                            hex_bytes, bits = line.rsplit(None, 1)
                        except ValueError:
                            # Entire line is trailing bits, '0.'
                            hex_bytes = None
                            bits = line
                    else:
                        hex_bytes = line
                        bits = None
                    # Handle regular hex bytes
                    if hex_bytes:
                        track_data.append(BitStream(hex=hex_bytes))
                    # Handle trailing binary item
                    if bits:
                        track_data.append(BitStream('0b' + bits[:-1]))
        # Add the track data. If no halftracks, store an empty track
        # as a placeholder.
        track_bin = BitStream().join(track_data)
        self += [(track_bin, density)]
        if not self.halftracks:
            self += [(BitStream(), -1)]

    def write_trkfile(self, filename):
        track_count = len(self)
        if not self.halftracks:
            track_count /= 2
        fp = open(filename, 'w')
        print >> fp, TRK_HEADER % {
            'DATE': time.ctime(),
            'ORIG_FILENAME': self.orig_filename,
            'TRACK_COUNT': track_count,
        }
        for i, x in enumerate(self):
            # Skip empty halftracks in output
            if not self.halftracks and (i & 1):
                continue
            print >> fp, 'track: %.1f' % ((float(i) + 2) / 2)
            data = '\n'.join(parse_track(x[0]))
            if data:
                print >> fp, data
                print >> fp, 'density: %d' % x[1]
            print >> fp
        fp.close()

    def rotate(self, track, amount):
        'Rotate the entire track right by the given amount of bits'
        track_data = self[track][0]
        if len(track_data) > 0:
            track_data.ror(amount)
        else:
            logging.warn('not rotating empty track %s', track)


def gcr_to_bytes(data):
    'Convert array of GCR symbols into bytes, 40 bits of 5:4 at a time'
    total_len = len(data)
    remainder = total_len % GCR_CHUNK_SIZE
    if remainder != 0:
        data = data + array.array('B', (GCR_CHUNK_SIZE - remainder) * [0])
    result = array.array('B')
    for i in xrange(0, len(data), GCR_CHUNK_SIZE):
        high_nibble = GCR_TABLE_HIGH[data[i] >> 3]
        low_nibble = GCR_TABLE_LOW[((data[i] << 2) | (data[i + 1] >> 6))
            & 0x1f]
        if high_nibble == -1 or low_nibble == -1:
            break
        result.append(high_nibble | low_nibble)
        high_nibble = GCR_TABLE_HIGH[(data[i + 1] >> 1) & 0x1f]
        low_nibble = GCR_TABLE_LOW[((data[i + 1] << 4) |
            (data[i + 2] >> 4)) & 0x1f]
        if high_nibble == -1 or low_nibble == -1:
            break
        result.append(high_nibble | low_nibble)
        high_nibble = GCR_TABLE_HIGH[((data[i + 2] << 1) |
            (data[i + 3] >> 7)) & 0x1f]
        low_nibble = GCR_TABLE_LOW[(data[i + 3] >> 2) & 0x1f]
        if high_nibble == -1 or low_nibble == -1:
            break
        result.append(high_nibble | low_nibble)
        high_nibble = GCR_TABLE_HIGH[((data[i + 3] << 3) |
            (data[i + 4] >> 5)) & 0x1f]
        low_nibble = GCR_TABLE_LOW[data[i + 4] & 0x1f]
        if high_nibble == -1 or low_nibble == -1:
            break
        result.append(high_nibble | low_nibble)
    return result

def parse_track(trk_stream):
    'Create an iterator that returns syncs and data bytes, both as strings'
    if len(trk_stream) == 0:
        return
    sync_length = 0
    chunk_itr = trk_stream.split(SYNC_SYMBOL)
    for chunk in chunk_itr:
        # If track starts with sync, first chunk is empty string
        if len(chunk) == 0:
            continue
        # Found more sync, so add to its length
        if chunk.all(1):
            sync_length += len(chunk)
        else:
            total_sync, data, trailing_bits = parse_data(sync_length, chunk)
            sync_length = 0
            if total_sync > 0:
                yield 'SYNC:%d' % total_sync
            # Return data in chunks of 10 hex bytes
            for chunk in chunk_iter(data, 10):
                gcr_str = ' '.join(chunk_iter(hexlify(chunk), 2))
                if len(trailing_bits) > 0 and len(chunk) < 10:
                    gcr_str += ' %s.' % trailing_bits.bin
                    trailing_bits = ''
                chunk_gcr = gcr_to_bytes(chunk)
                bytes_str = ' '.join(chunk_iter(hexlify(chunk_gcr), 2))
                yield '%-29s  # %s' % (gcr_str, bytes_str)
            # If we still have trailing bits, send that out too
            if len(trailing_bits) > 0:
                yield '%s.' % trailing_bits.bin
    # If we ended in a sync, return it too
    if sync_length != 0:
        yield 'SYNC:%d' % sync_length

def parse_data(sync_prefix, data):
    '''Finish calculating sync length and then return the following data
    as an array of bytes.
    '''
    # Find end of sync by looking for first 0 bit
    zero_offset = data.find('0b0')
    # No 0 bit, all sync
    if not zero_offset:
        return sync_prefix + len(data), '', ''
    # Compute total sync length in number of bits and the chunk of data itself
    total_sync = sync_prefix + data.pos
    remainder = (len(data) - data.pos) % 8
    if remainder:
        main_data = array.array('B', data[data.pos:-remainder].tobytes())
        trailing_bits = data[-remainder:]
    else:
        main_data = array.array('B', data[data.pos:].tobytes())
        trailing_bits = ''
    return total_sync, main_data, trailing_bits

def chunk_iter(buf, n):
    'Create an iterator that returns chunks of the given size from buf'
    return (buf[i:i+n] for i in xrange(0, len(buf), n))

def parse_density_map(data):
    '''Convert a byte-packed stream into a set of density values, 0-3.
    After we get the values for every byte, we compress them into
    ranges since there are often sequences of the same density value.
    TODO: implement/fix this.
    '''
    stream = BitStream(bytes=data)
    density_list = []
    density_values = [x.uint for x in stream.cut(2)]
    last_offset = 0
    last_density = density_values[0]
    for i, density in enumerate(density_values[1:]):
        if density != last_density:
            density_list += [(last_offset, i, last_density)]
            last_offset = offset
            last_density = density
    density_list += [(last_offset, i, last_density)]
    return density_list

def trk_data_lines(fp):
    for line in fp:
        line = line.split('#')[0].strip()
        if line:
            yield line

def get_kv(key, line):
    k, v = [x.strip() for x in line.split(':', 1)]
    if k != key:
        raise ValueError('improper config item "%s" in trk file' % k)
    return v

def read_process(input_file, output_file, transform=None):
    gcr_obj = GCRFile(input_file)
    # Modify the image with the requested operation
    if transform:
        tracks = [int(x * 2) for x in transform.tracks]
        for t in xrange(tracks[0], tracks[1] + 1):
            track = float(t) / 2
            if transform.rotate:
                gcr_obj.rotate(track, transform.rotate)
    gcr_obj.write(output_file)

def main():
    class ParseTrack(argparse.Action):
        'Convert one or two tracks into floating point values'
        def __call__(self, parser, namespace, values, option_string=None):
            tracks = [float(x) for x in values.split('-')]
            if len(tracks) > 2:
                raise ValueError('only one or two tracks allowed')
            elif len(tracks) == 1:
                tracks.append(tracks[0])
            setattr(namespace, self.dest, tracks)
    parser = argparse.ArgumentParser(description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-o', '--overwrite', action='store_true',
        help='overwrite input file as output after processing via flags')
    parser.add_argument('-t', '--tracks', action=ParseTrack,
        help='start through end (half)track to operate on. Example: -t 1-1.5')
    parser.add_argument('-r', '--rotate', type=int,
        help='rotate the specified track(s) right by N bits')
    parser.add_argument('input_file',
        help='file to be read (.trk or .g64 formats)')
    parser.add_argument('output_file', nargs='?',
        help='file to be written (.trk or .g64 formats)')
    args = parser.parse_args()
    # If we're doing a transformation, verify all options are present
    if args.tracks or args.rotate:
        if not (args.tracks and args.rotate):
            raise ValueError(
    'must specify both an action and track range if either option is present')
        transform = args
    else:
        transform = None
    if args.overwrite:
        if not args.output_file:
            args.output_file = args.input_file
    else:
        if not args.output_file:
            raise ValueError(
                'specify an output file or --overwrite (be careful!)')
        if os.path.isfile(args.output_file):
            raise ValueError(
                'output file already exists, rename or specify --overwrite')
    # Now choose and perform processing based on args
    read_process(args.input_file, args.output_file, transform=transform)

if __name__ == '__main__':
    if not PROFILING:
        main()
    else:
        import cProfile
        cProfile.run('main()', 'nibprof')
        import pstats
        p = pstats.Stats('nibprof')
        p.strip_dirs().sort_stats(-1).print_stats()
